/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *    Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: hostnames.c,v 1.6 2000/04/29 17:10:48 jordan Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "common-stuff.h"

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#else
# ifdef HAVE_SOCKET_H
#  include <socket.h>
# endif
#endif
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
# include <netdb.h>
#endif
#ifdef HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif
#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#include "hostnames.h"

/* ---------------------------------------------------------------------- *
 *                     private helpers                                    *
 * ---------------------------------------------------------------------- */

/* should be initialized in a threaded environment */
static const char *
get_myself
  (void)
{
  /* hold name for localhost */
  static char *myself;

  if (myself == 0) { 
    char buf [1024] ;
    if (gethostname (buf, sizeof (buf)) == 0 && buf [0])
      myself = XSTRDUP (buf) ;
  }
  
  return myself ;
}

/* upon parse error: set *is_error to 1 and return 0 */
static unsigned long
is_ip_pattern
  (const char *s, int *is_error)
{
  unsigned long result = 0 ;
  int a ;

  if ((a = strtol (s, (char**)&s, 10)) < 0 || a > 255 || s [0] != '.')
    goto parse_error;
  result = a ;
  
  if ((a = strtol (s+1, (char**)&s, 10)) < 0 || a > 255 || s [0] != '.')
    goto parse_error;
  result <<= 8 ;
  result  += a ;

  if ((a = strtol (s+1, (char**)&s, 10)) < 0 || a > 255 || s [0] != '.')
    goto parse_error;
  result <<= 8 ;
  result  += a ;

  if ((a = strtol (s+1, (char**)&s, 10)) < 0 || a > 255 || s [0] != '\0')
    goto parse_error;
  result <<= 8 ;
  result  += a ;
  
  return result ;

 parse_error:
  if (is_error != 0)
    *is_error = 1 ;
  return 0;
}

/* ---------------------------------------------------------------------- *
 *                     compare host names                                 *
 * ---------------------------------------------------------------------- */

const char *
get_my_host_name
   (void)
{
  return get_myself () ;
}

/* returns sort of a canonical host name */
const char *
get_host_ipaddress
   (const char *name)
{
  struct hostent *h;
  struct in_addr in;
  
  if (name == 0 && (name = get_myself ()) == 0)
    return 0 ;
  
  if (is_ip_pattern (name, 0))
    return name ;
  
  if ((h = gethostbyname (name)) == 0)
    return 0 ;

  /* convert to ip address */
  memcpy (&in.s_addr, *h->h_addr_list, h->h_length); 

  return inet_ntoa (in);
}

int
hostnomatch
  (const char *x,
   const char *y,
   size_t    len)
{
  return -1 ;
}


int
hostcmp
  (const char    *ntry,
   const char *pattern,
   size_t          len)
{
  char *buf, *p;
  struct hostent *h;
  struct in_addr in_ntry;

  /* quick check for impossible hostnames */
  if (*pattern == '#' || *pattern == '@')
    return -1 ;

  /* this is easy, is it not ? */
  if (len <= 0) {
    if (strcasecmp (pattern, ntry) == 0)
      return 0;   
    len = strlen (pattern) ;
  } else {
    if (strncasecmp (pattern, ntry, len) == 0)
      return 0;
  }
  
  /* check, whether the entry is pattern like (speed up common cases) */
  if (strchr (ntry, '/') != 0 || strchr (ntry, '+') != 0)
    return -1;
  
  /* resolve entry address numerically, ignore errors */
  p = 0;
  if (isdigit (ntry [0]) &&
      ((in_ntry.s_addr = strtol (ntry, &p, 10)) == LONG_MIN ||
       in_ntry.s_addr == LONG_MAX) &&
      errno == ERANGE ||  p == 0 || p [0] != '\0')
    p = 0 ;  /* no numeric address notation */
  
  /* resolve the entry address */
  if (p == 0) {
    if ((h = gethostbyname (ntry)) == 0)
      return -1 ;
    memcpy (&in_ntry.s_addr, *h->h_addr_list, h->h_length); 
  }
  
  /* now, the general case: cut off trailing garbage from the pattern */
  buf = ALLOCA (len + 1);
  memcpy (buf, pattern, len) ;
  buf [len] = '\0' ;

  /* for each segment of the pattern list (delimited by +) do */
  if ((pattern = strtok (buf, "+")) != 0)
    do {
      int n ;
      unsigned long mask = -1, ntry_addr ;
      struct in_addr in1 ;
      
      /* check for <entry-string> ::= <ipnet>/<size> notation */
      if ((p = strchr (pattern, '/')) != 0) {
	p [0] = '\0' ;
	/* /gen.eral.net.mask notation ? */
	if (strchr (p+1, '.') != 0) {
	  if ((mask = is_ip_pattern (p+1, (n=0, &n))) == 0 && n)
	    /* syntax error */
	    break ;
	} else {
	  /* /<number-of-leading-bits> notation */
	  if ((mask = strtol (p+1, &p, 10)) < 0 || mask > 32 || p [0] != '\0')
	    /* improper address notation -- syntax error */
	    break ;
	  if (mask)
	    /* convert to bit notation */
	    mask = (mask < 32 ? (((((long)1)<<mask)-1)<<(32-mask)) : -1);
	}
	/* convert to network notation */
	mask = htonl (mask);
      }
    
      /* resolve current pattern address, ignore errors */
      p = 0;
      if (isdigit (pattern [0]) &&
	  ((in1.s_addr = strtol (pattern, &p, 10)) == LONG_MIN ||
	   in1.s_addr == LONG_MAX) &&
	  errno == ERANGE ||  p == 0 || p [0] != '\0')
	p = 0 ;  /* no numeric address notation */

      /* resolve host name, otherwise */
      if (p == 0) {
	if ((h = gethostbyname (pattern)) == 0) continue ;
	memcpy (&in1.s_addr, *h->h_addr_list, h->h_length);
      }

      /* mask off host bits => network match */
      ntry_addr = (in_ntry.s_addr & mask) ;

      /* for ipv4, in.s_addr is a 32 bit type, eg. unsigned int */
      if (in1.s_addr != ntry_addr) continue ;
      
      /* clean up: entry is in pattern */
      DEALLOCA (buf) ;
      return 0 ;
      
      /* get next segment */
    } while ((pattern = strtok (0, "+")) != 0);

  /* clean up: no match */
  DEALLOCA (buf) ;
  return -1 ;
}
